import 'package:flutter/material.dart';

/// A service for navigation.
///
/// This is automatically registered as a singleton.
///
/// Views that will be navigated to must be registered using
/// [ServiceRegistry.registerView].
abstract class NavigationService {
  /// The key to be used to determine what [Navigator] to use during navigation.
  ///
  /// Assign this to your [MaterialApp.navigatorKey].
  static final navigatorKey = GlobalKey<NavigatorState>();

  /// Pops the current route.
  ///
  /// Optionally pass a [result] when the route was popped.
  void pop([Map<String, Object> result]);

  /// Pops the current route until [viewName].
  void popUntil(String viewName);

  /// Pushes a new route.
  ///
  /// Optionally pass [parameters] to the new route.
  Future<Map<String, Object>> push(String viewName, [Map<String, Object> parameters]);

  /// Pushes a new route, and removes all previous route until [removeUntil].
  ///
  /// Optionally pass [parameters] to the new route.
  Future<Map<String, Object>> pushAndRemoveUntil(
    String viewName,
    String removeUntil, [
    Map<String, Object> parameters,
  ]);

  /// Pushes a new route as the root of the navigation stack.
  ///
  /// Optionally pass [parameters] to the new route.
  void pushToNewRoot(String viewName, [Map<String, Object> parameters]);

  /// Pushes a route to be presented modally.
  ///
  /// Optionally pass [parameters] to the new route.
  Future<Map<String, Object>> pushModal(String viewName, [Map<String, Object> parameters]);

  /// Pushes a new route as a replacement of the current one.
  ///
  /// Optionally pass [parameters] to the new route.
  Future<Map<String, Object>> pushReplacement(String viewName, [Map<String, Object> parameters]);
}
