import 'package:arch/src/application.dart';
import 'package:arch/src/dialog.dart';
import 'package:arch/src/initializable.dart';
import 'package:arch/src/view_model.dart';
import 'package:flutter/material.dart';
import 'package:meta/meta.dart';
import 'package:provider/provider.dart';

/// A view that is stateless and has a view model.
///
/// [TViewModel] is automatically resolved, and initializes it when navigated to.
abstract class ViewBase<TViewModel extends ViewModelBase> extends StatelessWidget {
  ViewBase({Key key}) : super(key: key);

  /// Describes the UI to be presented by this view.
  ///
  /// Called whenever the view model changes.
  Widget buildView(BuildContext context, TViewModel viewModel);

  @nonVirtual
  @override
  Widget build(BuildContext context) {
    return ListenableProvider<TViewModel>(
      create: (_) {
        assert(Application.serviceLocator.canResolve<TViewModel>(),
            'Cannot find view model $TViewModel for ${this.runtimeType}. Did you registered it?');

        final viewModel = Application.serviceLocator.resolve<TViewModel>();
        final parameters = ModalRoute.of(context)?.settings?.arguments as Map<String, Object>;
        viewModel.init(parameters);
        didInitViewModel(viewModel);
        return viewModel;
      },
      child: Consumer<TViewModel>(builder: (context, viewModel, _) {
        return buildView(context, viewModel);
      }),
      dispose: (_, viewModel) => dispose(viewModel),
    );
  }

  /// Called after initializing the view model.
  ///
  /// Set objects here that depend on the view model after it was initialized.
  void didInitViewModel(TViewModel viewModel) {}

  /// Free resources in this object before removing it from the widget tree.
  ///
  /// If you override this, make sure to end your method with a call to super.dispose(),
  /// because this will dispose the view model after all disposal in this view is finished.
  @mustCallSuper
  void dispose(TViewModel viewModel) {
    viewModel.dispose();
  }
}

/// A view that is presented as a dialog. It is stateless and can be navigated to.
abstract class DialogBase<TDialogModel extends DialogModelBase> extends StatelessWidget {
  DialogBase({Key key}) : super(key: key);

  /// Describes the UI to be presented by this dialog.
  ///
  /// Called whenever the dialog model changes.
  Widget buildDialog(BuildContext context, TDialogModel dialogModel);

  @nonVirtual
  @override
  Widget build(BuildContext context) {
    return ListenableProvider<TDialogModel>(
      create: (_) {
        assert(Application.serviceLocator.canResolve<TDialogModel>(),
            'Cannot find dialog model $TDialogModel for ${this.runtimeType}. Did you registered it?');

        final dialogModel = Application.serviceLocator.resolve<TDialogModel>();
        final parameters = DialogRoute.of(context)?.settings?.arguments as Map<String, Object>;
        dialogModel.init(parameters);
        didInitDialogModel(dialogModel);
        return dialogModel;
      },
      child: Consumer<TDialogModel>(builder: (context, dialogModel, _) => buildDialog(context, dialogModel)),
      dispose: (_, dialogModel) => dispose(dialogModel),
    );
  }

  /// Called after initializing the dialog model.
  ///
  /// Set objects here that depend on the dialog model after it was initialized.
  void didInitDialogModel(TDialogModel dialogModel) {}

  /// Free resources in this object before removing it from the widget tree.
  ///
  /// If you override this, make sure to end your method with a call to super.dispose(),
  /// because this will dispose the dialog model after all disposal in this dialog is finished.
  @mustCallSuper
  void dispose(TDialogModel dialogModel) {
    dialogModel.dispose();
  }
}

/// A view's state that has a view model.
///
/// [TViewModel] is automatically resolved, and initializes it when navigated to.
abstract class ViewStateBase<TWidget extends StatefulWidget, TViewModel extends ViewModelBase>
    extends ViewModelBoundState<TWidget, TViewModel> with ViewModelAutoResolver, ViewModelNavigationInitializer {
  /// Describes the UI to be presented by this state.
  ///
  /// Called whenever the view model changes.
  Widget buildView(BuildContext context);

  @nonVirtual
  @override
  Widget build(BuildContext context) {
    return ListenableProvider<TViewModel>(
      create: (_) => viewModel,
      child: Consumer<TViewModel>(builder: (context, _, __) => buildView(context)),
    );
  }
}

/// A view's state that has a view model.
///
/// You will need to provide the [TViewModel] for this state.
abstract class ViewModelBoundState<TWidget extends StatefulWidget, TViewModel extends ChangeNotifier>
    extends State<TWidget> {
  TViewModel get viewModel;

  /// Free resources in this object before removing it from the widget tree.
  ///
  /// If you override this, make sure to end your method with a call to super.dispose(),
  /// because this will dispose the view model after all disposal in this state is finished.
  @mustCallSuper
  @override
  void dispose() {
    assert(viewModel != null);

    viewModel.dispose();
    super.dispose();
  }
}

/// A mixin for [ViewModelBoundState] that automatically resolves [TViewModel].
///
/// [TViewModel] must be registered in [Initializer.registerTypes].
mixin ViewModelAutoResolver<TWidget extends StatefulWidget, TViewModel extends ChangeNotifier>
    on ViewModelBoundState<TWidget, TViewModel> {
  TViewModel _viewModel;

  @nonVirtual
  TViewModel get viewModel => _viewModel;

  @override
  @mustCallSuper
  void initState() {
    super.initState();
    assert(Application.serviceLocator.canResolve<TViewModel>(),
        'Cannot find dialog model $TViewModel for ${this.runtimeType}. Did you registered it?');

    _viewModel = Application.serviceLocator.resolve<TViewModel>();
  }
}

/// A mixin for [ViewModelBoundState] that initializes [TViewModel] when
/// this state was navigated to.
///
/// Initialization occurs during [State.didChangeDependencies], and happens
/// only once. It will get the [RouteSettings.arguments] and pass it
/// to [TViewModel].
///
/// It is needed that [TViewModel] should implement [Initializable], and the
/// passed [RouteSettings.arguments] should be of type [Map<String, Object>].
mixin ViewModelNavigationInitializer<TWidget extends StatefulWidget, TViewModel extends ChangeNotifier>
    on ViewModelBoundState<TWidget, TViewModel> {
  bool _initializeChecked = false;

  /// Called after initializing the view model.
  ///
  /// Set objects here that depend on the view model after it was initialized.
  void didInitViewModel() {}

  @override
  @mustCallSuper
  void didChangeDependencies() {
    super.didChangeDependencies();

    if (!_initializeChecked) {
      assert(viewModel != null);
      assert(viewModel is Initializable);

      final parameters = ModalRoute.of(context)?.settings?.arguments;
      final initializable = viewModel as Initializable;
      initializable.init(parameters);
      didInitViewModel();
      _initializeChecked = true;
    }
  }
}

/// A mixin for [ViewModelBoundState] that initializes [TViewModel]
/// when [State.initState] was called.
///
/// Initialization, like [State.initState], happens only once.
///
/// It is needed that the view model should implement [Initializable].
mixin ViewModelInitializer<TWidget extends StatefulWidget, TViewModel extends ChangeNotifier>
    on ViewModelBoundState<TWidget, TViewModel> {
  /// Called after initializing the view model.
  ///
  /// Set objects here that depend on the view model after it was initialized.
  void didInitViewModel() {}

  @override
  @mustCallSuper
  void initState() {
    super.initState();

    assert(viewModel != null);
    assert(viewModel is Initializable);

    final initializable = viewModel as Initializable;
    initializable.init();
    didInitViewModel();
  }
}
